use core::panic;
use std::net::{IpAddr, Ipv4Addr, SocketAddr};

use axum::extract::ConnectInfo;
use axum::http::header::FORWARDED;
use axum::http::HeaderMap;
use axum::{extract::State, http::StatusCode, routing::get, Json, Router};
use rand::random;
use thiserror::Error;

use tracing::{debug, info, instrument};

mod types;
use types::*;

#[derive(Error, Debug)]
pub enum HttpServError {
    #[error("OS Error: {0}")]
    IoError(#[from] std::io::Error),
}

use crate::conf::Config;
use crate::db::DB;

use crate::types::*;

use rfc7239::parse;

#[derive(Clone)]
pub struct AppState {
    db: DB,
    conf: Config,
}

type Answer = (u64, Vec<Entry>);
const EMPTY_ANSWER: Json<Answer> = Json((0, Vec::new()));

/// Start to listen for incoming http connections.
pub async fn run_web_api(db: DB, conf: Config) {
    match run_web_api_failable(db, conf).await {
        Ok(_) => unreachable!(),
        Err(e) => panic!("Big failure! {:?}", e),
    }
}

async fn run_web_api_failable(db: DB, conf: Config) -> Result<(), HttpServError> {
    info!("Binding to {}", conf.http_ip4);
    let listener = tokio::net::TcpListener::bind(&conf.http_ip4).await?;
    let app = Router::new()
        .route("/last", get(query_last_n))
        .route("/all", get(query_all))
        .with_state(AppState { db, conf });
    axum::serve(
        listener,
        app.into_make_service_with_connect_info::<SocketAddr>(),
    )
    .await?;
    Ok(())
}

/// Checks for an API key, and then checks if an ip is allowed.
async fn allowed(conf: &Config, db: &DB, headers: &HeaderMap, ip: &IpAddr) -> bool {
    if let Some(key) = headers.get("x-api-key") {
        if let Ok(key_str) = key.to_str() {
            match db.get_key(key_str).await {
                Ok(_) => return true,
                Err(e) => debug!("AuthError: {e:?}"),
            }
        }
    }

    match ip {
        std::net::IpAddr::V4(ip) => {
            let trueip = get_true_ip(conf, headers, ip);
            if !conf.whitelist.contains(&trueip) {
                return false;
            }
        }
        std::net::IpAddr::V6(_) => todo!("IPV6 still to do!!"),
    }

    true
}

fn get_true_ip(conf: &Config, headers: &HeaderMap, ip: &Ipv4Addr) -> Ipv4Addr {
    if conf.trusted_proxies.contains(ip) {
        if let Some(forwarded_for) = headers.get(FORWARDED) {
            if let Ok(forwarded_for_str) = forwarded_for.to_str() {
                debug!("forwarded_for_headers:{forwarded_for:?}");
                if let Some(addr) = parse_forwarded_for(forwarded_for_str) {
                    return addr;
                };
            };
        }
    }

    *ip
}

/// Wrapping our forwarded_for library
fn parse_forwarded_for(header_val: &str) -> Option<Ipv4Addr> {
    for node_result in parse(header_val) {
        let node = node_result.ok()?;
        if let Some(forwarded_for) = node.forwarded_for {
            debug!("Forwarded by {}", forwarded_for)
            // TODO: boh?
        }
    }
    None
}

#[instrument(skip(state))]
async fn query_last_n(
    ConnectInfo(addr): ConnectInfo<SocketAddr>,
    headers: HeaderMap,
    State(state): State<AppState>,
    Json(payload): Json<QueryN>,
) -> (StatusCode, Json<Answer>) {
    let n = payload.n.unwrap_or(1);
    debug!("Headers: {headers:?} Request: {payload:?} from {addr}");

    if !state.conf.public_sets.contains(&payload.name)
        && !allowed(&state.conf, &state.db, &headers, &addr.ip()).await
    {
        return (StatusCode::NETWORK_AUTHENTICATION_REQUIRED, EMPTY_ANSWER);
    };

    match state.db.get_last_n(&payload.name, n).await {
        Ok(answers) => return (StatusCode::OK, Json((answers.len() as u64, answers))),
        Err(e) => {
            debug!("{e:?}");
            return (StatusCode::NOT_FOUND, EMPTY_ANSWER);
        }
    }
}

#[instrument(skip(state))]
async fn query_all(
    ConnectInfo(addr): ConnectInfo<SocketAddr>,
    headers: HeaderMap,
    State(state): State<AppState>,
    Json(payload): Json<QueryAll>,
) -> (StatusCode, Json<Answer>) {
    debug!("Headers: {headers:?} Request: {payload:?} from {addr}");

    // If the queried datased is NOT public, check for
    // ip whitelisting.
    if !state.conf.public_sets.contains(&payload.name)
        && !allowed(&state.conf, &state.db, &headers, &addr.ip()).await
    {
        return (StatusCode::NETWORK_AUTHENTICATION_REQUIRED, EMPTY_ANSWER);
    };

    match (payload.from, payload.to) {
        (None, None) => match state.db.get_all_values(&payload.name).await {
            Ok(answers) => (StatusCode::OK, Json((answers.len() as u64, answers))),
            Err(e) => {
                debug!("{e:?}",);
                (StatusCode::NO_CONTENT, EMPTY_ANSWER)
            }
        },
        (None, Some(end)) => {
            match state
                .db
                .get_from_to_date(&payload.name, DATE_MIN, end)
                .await
            {
                Ok(answers) => (StatusCode::OK, Json((answers.len() as u64, answers))),
                Err(e) => {
                    debug!("{e:?}");
                    (StatusCode::NO_CONTENT, EMPTY_ANSWER)
                }
            }
        }
        (Some(start), None) => {
            match state
                .db
                .get_from_to_date(&payload.name, start, DATE_MAX)
                .await
            {
                Ok(answers) => (StatusCode::OK, Json((answers.len() as u64, answers))),
                Err(e) => {
                    debug!("{e:?}");
                    (StatusCode::NO_CONTENT, EMPTY_ANSWER)
                }
            }
        }
        (Some(start), Some(end)) => {
            match state.db.get_from_to_date(&payload.name, start, end).await {
                Ok(answers) => (StatusCode::OK, Json((answers.len() as u64, answers))),
                Err(e) => {
                    debug!("{e:?}");
                    (StatusCode::NO_CONTENT, EMPTY_ANSWER)
                }
            }
        }
    }
}

/// Generates random API key. It's like a long, strong password. I guess.
pub async fn generate_api_key() -> String {
    use rand::Rng;
    const CHARSET: &[u8] = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ\
                            abcdefghijklmnopqrstuvwxyz\
                            0123456789";
    const API_KEY_LEN: usize = 30;
    let mut rng = rand::thread_rng();

    let api_key: String = (0..API_KEY_LEN)
        .map(|_| {
            let idx = rng.gen_range(0..CHARSET.len());
            CHARSET[idx] as char
        })
        .collect();

    api_key
}
