use std::{
    io::{self, Write},
    process::exit,
};

use crate::api::{self, run_web_api};

use crate::tcp;
use tcp::listen_api;
use tokio::task;
use tracing::{debug, info};

use crate::cli::*;
use crate::conf::Config;
use crate::db::DB;
use crate::types;

use types::{DATE_MAX, DATE_MIN};

pub async fn dispatch(args: Args, conf: &Config) {
    match args.command {
        Command::Run => {
            // Initializing database
            let db = DB::new(&conf.db).await.unwrap();
            let local = task::LocalSet::new();
            // both listen_api and run_web_api do their error reporting
            // through logging.
            // so it needs not to be implemented here.
            local.spawn_local(listen_api(db.clone(), conf.clone()));
            local.spawn_local(run_web_api(db.clone(), conf.clone()));
            local.await
        }
        Command::ShowConf => {
            println!("{conf}");
        }
        Command::Delete {
            name,
            from,
            to,
            force_yes,
        } => {
            let db = DB::new(&conf.db).await.unwrap();

            if !force_yes {
                if let Err(e) =
                    ask_for_confirmation(&format!("Deleting dataset {name}. Are you sure?")).await
                {
                    panic!("Could not print to screen? {e:?}");
                };
            }

            match (from, to) {
                (None, None) => db.delete_all(&name).await.unwrap(),
                (None, Some(end)) => db.delete_from_to_date(&name, DATE_MAX, end).await.unwrap(),
                (Some(start), None) => db
                    .delete_from_to_date(&name, start, DATE_MIN)
                    .await
                    .unwrap(),
                (Some(start), Some(end)) => {
                    db.delete_from_to_date(&name, start, end).await.unwrap()
                }
            }
        }
        Command::ShowData => todo!(),
        Command::Info => todo!(),
        Command::Api { subcommand } => {
            let db = DB::new(&conf.db).await.unwrap();
            match subcommand {
                ApiCommand::Generate { comment } => {
                    let key = api::generate_api_key().await;
                    let res = comment.unwrap_or(String::new());
                    match db.add_key(&key, &res).await {
                        Ok(_) => {
                            info!("Added api key to database. Copy your key:");
                            println!("{key}");
                        }
                        Err(e) => {
                            debug!("{e:?}");
                            println!("Error generating/adding key.")
                        }
                    };
                }

                ApiCommand::List => match db.get_all_keys().await {
                    Ok(keys) => {
                        for key in keys {
                            println!(
                                "id: {} key: {} comment: {}",
                                key.key_id, key.key, key.comment
                            );
                        }
                    }
                    Err(e) => {
                        debug!("{e:?}");
                        println!("Error getting key list.");
                    }
                },
                ApiCommand::Delete { id } => match db.delete_key(id).await {
                    Ok(_) => {
                        println!("Deleted key with id: {id}.");
                    }
                    Err(e) => {
                        debug!("{e:?}");
                        println!("Could not delete or find key.");
                    }
                },
            }
        }
    }
}

async fn ask_for_confirmation(msg: &str) -> Result<(), io::Error> {
    print!("{msg} y/n: ");
    let mut answer = String::new();
    io::stdout().flush()?;
    io::stdin().read_line(&mut answer)?;
    if answer.trim() != "y" {
        println!("Canceling operation.");
        exit(1);
    };
    Ok(())
}
