use serde::{Deserialize, Serialize};
use std::{fmt::Display, net::Ipv4Addr};
use thiserror::Error;
use tokio::{fs::File, io::AsyncReadExt};

#[derive(Clone, Serialize, Deserialize)]
pub struct Config {
    pub tcp_ip4: String,
    pub tcp_ip6: Option<String>,
    pub http_ip4: String,
    pub db: String,
    pub whitelist: Vec<Ipv4Addr>,
    pub trusted_proxies: Vec<Ipv4Addr>,
    pub public_sets: Vec<String>,
}

#[derive(Error, Debug)]
pub enum ConfError {
    #[error("File Error: {0}")]
    FileError(#[from] tokio::io::Error),
    #[error("Parse Error: {0}")]
    /// TODO: Prettify parse errors
    ParseErrpr(#[from] toml::de::Error),
}

impl Default for Config {
    fn default() -> Self {
        Self {
            tcp_ip4: String::from("0.0.0.0:58008"),
            tcp_ip6: None,
            http_ip4: String::from("0.0.0.0:3000"),
            db: String::from("sqlite:file:db?cache=shared"),
            whitelist: vec!["127.0.0.1".parse().unwrap()],
            trusted_proxies: vec![],
            public_sets: vec![],
        }
    }
}

impl Display for Config {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            toml::to_string(self).expect("Unrecoverable error: Config can't be serialized to toml")
        )
    }
}

impl Config {
    pub async fn new(path: &str) -> Result<Self, ConfError> {
        let mut conffile = File::open(path).await?;
        let mut contents = String::new();
        conffile.read_to_string(&mut contents).await?;
        let conf = toml::from_str(&contents)?;
        Ok(conf)
    }
}
