use sqlx::{
    sqlite::{SqliteConnectOptions, SqlitePoolOptions},
    Pool, Sqlite,
};
use std::str::FromStr;
use tracing::{info, instrument};

use thiserror::Error;

use crate::types::*;
// TODO: Support for other databases
#[derive(Clone)]
pub struct DB(Pool<Sqlite>);

#[derive(Error, Debug)]
pub enum DBError {
    #[error("Sql error: {0}")]
    SqlError(#[from] sqlx::error::Error),
}

const WHERE_LAST_N: &str = "WHERE data_id = (SELECT data_id FROM data_tables WHERE name = ?1)
            LIMIT ?2";

const WHERE_ALL: &str = "WHERE
            data_id = (SELECT data_id FROM data_tables WHERE name = ?1)
            ";

const WHERE_FROM_TO: &str = "WHERE
            data_id = (SELECT data_id FROM data_tables WHERE name = ?1) AND
            date BETWEEN ?2 AND ?3
            ";

impl DB {
    /// Creates a new database, currently Sqlite3.
    #[instrument]
    pub async fn new(db_url: &str) -> Result<Self, DBError> {
        info!("Connecting to the database at: {db_url}");
        let conops = SqliteConnectOptions::from_str(db_url)?.create_if_missing(true);

        let pool = SqlitePoolOptions::new()
            .max_connections(10)
            .connect_with(conops)
            .await?;

        let db = DB(pool);
        db.create_tables().await?;
        Ok(db)
    }

    /// Creates, if they don't exist, two database tables.
    ///
    /// The first table stores name of data sets, togheter with an id. (dataset_id : dataset_name)
    /// The second table stores every message received. (dataset_id : value : date)
    ///
    // TODO: This is not immensely efficient.
    #[instrument(skip_all)]
    async fn create_tables(&self) -> Result<(), DBError> {
        info!("Checking if tables exists and creating them.");

        sqlx::query(
            "CREATE TABLE IF NOT EXISTS data_tables (
                data_id INTEGER PRIMARY KEY NOT NULL,
                name TEXT NOT NULL,
                unique(name)
                )",
        )
        .execute(&(self.0))
        .await?;

        sqlx::query(
            "CREATE TABLE IF NOT EXISTS api_keys (
                key_id INTEGER PRIMARY KEY NOT NULL,
                key TEXT NOT NULL,
                comment TEXT,
                unique(key)
                )",
        )
        .execute(&(self.0))
        .await?;

        sqlx::query(
            "CREATE TABLE IF NOT EXISTS data (
                data_id INTEGER NOT NULL,
                value TEXT NOT NULL,
                date INTEGER NOT NULL,
                foreign key (data_id) references data_tables (data_id)
                )",
        )
        .execute(&(self.0))
        .await?;
        Ok(())
    }

    // TODO: Hash keys!

    /// Adds API key.
    pub async fn add_key(&self, api_key: &str, comment: &str) -> Result<(), DBError> {
        sqlx::query("INSERT INTO api_keys ( key, comment ) VALUES ( ?1, ?2 )")
            .bind(api_key)
            .bind(comment)
            .execute(&(self.0))
            .await?;
        Ok(())
    }

    /// Deletes an API key.
    pub async fn delete_key(&self, key_id: i64) -> Result<(), DBError> {
        sqlx::query("DELETE FROM api_keys WHERE key_id=?1")
            .bind(key_id)
            .execute(&(self.0))
            .await?;
        Ok(())
    }

    /// Gets API key.
    pub async fn get_key(&self, api_key: &str) -> Result<ApiKey, DBError> {
        let key = sqlx::query_as::<_, ApiKey>(&format!(
            "SELECT key_id, key, comment FROM api_keys WHERE key=?1;"
        ))
        .bind(api_key)
        .fetch_one(&(self.0))
        .await?;
        Ok(key)
    }

    /// Gets all API keys.
    pub async fn get_all_keys(&self) -> Result<Vec<ApiKey>, DBError> {
        let keys = sqlx::query_as::<_, ApiKey>("SELECT key_id, key, comment FROM api_keys")
            .fetch_all(&(self.0))
            .await?;
        Ok(keys)
    }

    /// Adds a certain value to the dataset <name>. This function performs lookup
    /// of datasets, eventul creation of dataset and then finally insertion.
    /// When repeatedly sending the same data, it is best to create an
    /// EntryWriter.
    //
    // TODO: EntryWriter link.
    //
    // Internally, this function initially performs a lookup, and if it fails
    // it creates a record and returns the newly created id.
    pub async fn write_entry(&self, name: &str, val: &str) -> Result<(), DBError> {
        let writer = EntryWriter::new(name, self).await?;
        writer.write_entry(val).await
    }

    /// Retrives the lastest <n> values from the dataset <name>.
    pub async fn get_last_n(&self, name: &str, n: usize) -> Result<Vec<Entry>, DBError> {
        let n_i = n as i64;

        let last_vals = sqlx::query_as::<_, Entry>(&format!(
            "SELECT value, date FROM data {WHERE_LAST_N} ORDER BY date DESC"
        ))
        .bind(name)
        .bind(n_i)
        .fetch_all(&(self.0))
        .await?;

        Ok(last_vals)
    }

    /// Deletes the lastest <n> values from the dataset <name>.
    pub async fn delete_last_n(&self, name: &str, n: usize) -> Result<(), DBError> {
        let n_i = n as i64;

        sqlx::query(&format!("DELETE FROM data {WHERE_LAST_N}"))
            .bind(name)
            .bind(n_i)
            .fetch_all(&(self.0))
            .await?;

        Ok(())
    }

    /// Retrives all values from the dataset <name>.
    pub async fn get_all_values(&self, name: &str) -> Result<Vec<Entry>, DBError> {
        let vals = sqlx::query_as::<_, Entry>(&format!(
            "SELECT value, date FROM data {WHERE_ALL} ORDER BY date DESC"
        ))
        .bind(name)
        .fetch_all(&(self.0))
        .await?;

        Ok(vals)
    }

    pub async fn delete_all(&self, name: &str) -> Result<(), DBError> {
        sqlx::query(&format!("DELETE FROM data {WHERE_ALL}"))
            .bind(name)
            .fetch_all(&(self.0))
            .await?;

        Ok(())
    }

    /// Retrives values from <start> to <end> date.
    pub async fn get_from_to_date(
        &self,
        name: &str,
        start: Date,
        end: Date,
    ) -> Result<Vec<Entry>, DBError> {
        let last_vals = sqlx::query_as::<_, Entry>(&format!(
            "SELECT value, date FROM data {WHERE_FROM_TO} ORDER BY date DESC"
        ))
        .bind(name)
        .bind(start)
        .bind(end)
        .fetch_all(&(self.0))
        .await?;

        Ok(last_vals)
    }

    /// Deletes values from <start> to <end> date.
    pub async fn delete_from_to_date(
        &self,
        name: &str,
        start: Date,
        end: Date,
    ) -> Result<(), DBError> {
        sqlx::query(&format!("DELETE FROM data {WHERE_FROM_TO}"))
            .bind(name)
            .bind(start)
            .bind(end)
            .fetch_all(&(self.0))
            .await?;

        Ok(())
    }
}

pub struct EntryWriter<'a> {
    db: &'a DB,
    entry_val: i64,
}

/// Specialized struct to write many values to the same entry.
impl<'a> EntryWriter<'a> {
    #[instrument(skip(db))]
    pub async fn new(name: &str, db: &'a DB) -> Result<Self, DBError> {
        #[allow(dead_code)]
        #[derive(sqlx::FromRow)]
        struct RowId {
            data_id: i64,
        }

        let query =
            sqlx::query_as::<_, RowId>("SELECT data_id FROM data_tables WHERE name = ?1 LIMIT 1")
                .bind(name)
                .fetch_one(&db.0)
                .await;

        let name_id = match query {
            Ok(o) => o.data_id,
            Err(_) => {
                let query_id = sqlx::query("INSERT INTO data_tables ( name ) VALUES ( ?1 )")
                    .bind(name)
                    .execute(&db.0)
                    .await?;
                query_id.last_insert_rowid()
            }
        };
        let entry_writer = Ok(EntryWriter {
            db,
            entry_val: name_id,
        });
        return entry_writer;
    }

    /// Writes a certain entry <val> with a known dataset <id>.
    pub async fn write_entry(&self, val: &str) -> Result<(), DBError> {
        sqlx::query(
            "INSERT INTO data ( data_id, value, date) VALUES ( ?1, ?2, strftime('%s','now'))",
        )
        .bind(self.entry_val)
        .bind(val)
        .execute(&(self.db.0))
        .await?;

        Ok(())
    }
}
