use thiserror::Error;
use tokio::{
    io::{AsyncBufReadExt, BufReader},
    net::{TcpListener, TcpStream},
    task,
};
use tracing::{debug, error, info, instrument};

use std::io;
use std::net::SocketAddr;

use crate::conf::Config;
use crate::db::{EntryWriter, DB};

#[derive(Error, Debug)]
pub enum TcpError {
    #[error("Sql error: {0}")]
    DBError(#[from] crate::db::DBError),
    #[error("I/O error: {0}")]
    IOerr(#[from] io::Error),
    #[error("No input")]
    NoData,
}

/// Start to listen for incoming tcp connections.
pub async fn listen_api(dbobj: DB, conf: Config) {
    match listen_api_failable(dbobj, conf).await {
        Ok(_) => unreachable!(),
        Err(e) => match e {
            TcpError::DBError(e) => debug!("{e:?}"),
            TcpError::IOerr(e) => debug!("{e:?}"),
            TcpError::NoData => unreachable!(),
        },
    }
}

#[instrument(skip_all, name = "tcp_api")]
pub async fn listen_api_failable(dbobj: DB, conf: Config) -> Result<(), TcpError> {
    info!("Binding to {}", &conf.tcp_ip4);
    let listener = TcpListener::bind(&conf.tcp_ip4).await?;

    loop {
        let (socket, addr) = listener.accept().await?;
        debug!("Incoming connection from {addr:?}");

        let dbtmp = dbobj.clone();
        task::spawn(async move { handle_connection(socket, &dbtmp, &addr).await });
    }
}

// This function shouldn't be called directly. Instead, use handle_connections.
// On success, it returns the dataset name and how many entries were sent.
async fn process_socket(socket: TcpStream, dbobj: &DB) -> Result<(String, i64), TcpError> {
    let reader = BufReader::new(socket);

    let mut lines = reader.lines();

    if let Ok(Some(line)) = lines.next_line().await {
        let service_name = line;

        let entry_writer = EntryWriter::new(&service_name, dbobj).await?;

        let mut n = 0;

        while let Ok(Some(line)) = lines.next_line().await {
            n += 1;
            entry_writer.write_entry(&line).await?;
        }
        Ok((service_name, n))
    } else {
        Err(TcpError::NoData)
    }
}

// Since process_socket is meant to be run with tokio::task(), this function
// handles error management and reporting. Pretty much everything here can fail without
// crashing the application.
#[instrument(skip_all, name = "tcp_api")]
async fn handle_connection(socket: TcpStream, dbobj: &DB, addr: &SocketAddr) {
    match process_socket(socket, dbobj).await {
        Ok((name, n)) => debug!("Wrote {n} messages with name {name} from {addr}"),
        Err(e) => match e {
            TcpError::DBError(e) => error!("Database error while talking to {addr}: {e:?}"),
            TcpError::IOerr(e) => error!("I/O error while talking to {addr}: {e:?}"),
            TcpError::NoData => debug!("No data came from {addr}"),
        },
    }
}
